# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"convert_file_to_json": "00-documentation.ipynb",
         "encode_file_as_utf8": "00-documentation.ipynb",
         "convert_nb_to_md": "00-documentation.ipynb",
         "junix.exporter.convert_file_to_json": "00-documentation.ipynb",
         "MyHTMLParser": "00-documentation.ipynb",
         "convert_df_to_md": "00-documentation.ipynb",
         "remove_empty_index_col": "00-documentation.ipynb",
         "extract_div_to_md_table": "00-documentation.ipynb",
         "extract_div_to_md_tables": "00-documentation.ipynb",
         "clean_md_file_tables": "00-documentation.ipynb",
         "get_substring_idxs": "00-documentation.ipynb",
         "clean_progress_bars": "00-documentation.ipynb",
         "correct_png_name": "00-documentation.ipynb",
         "specify_nb_in_img_fp": "00-documentation.ipynb",
         "get_nb_naive_png_names": "00-documentation.ipynb",
         "get_filename_correction_map": "00-documentation.ipynb",
         "clean_md_text_img_fps": "00-documentation.ipynb",
         "clean_md_files_img_fps": "00-documentation.ipynb",
         "convert_and_clean_nb_to_md": "00-documentation.ipynb",
         "RequestError": "01-utils.ipynb",
         "check_status": "01-utils.ipynb",
         "check_capping": "01-utils.ipynb",
         "expand_cols": "01-utils.ipynb",
         "parse_xml_response": "01-utils.ipynb",
         "dt_rng_to_SPs": "01-utils.ipynb",
         "parse_local_datetime": "01-utils.ipynb",
         "init_spec": "02-spec-gen.ipynb",
         "load_endpoints_df": "02-spec-gen.ipynb",
         "get_endpoint_single_attr": "02-spec-gen.ipynb",
         "init_stream_dict": "02-spec-gen.ipynb",
         "add_params_to_stream_dict": "02-spec-gen.ipynb",
         "add_streams_to_spec": "02-spec-gen.ipynb",
         "construct_spec": "02-spec-gen.ipynb",
         "save_spec": "02-spec-gen.ipynb",
         "load_API_yaml": "02-spec-gen.ipynb",
         "clean_path_name": "03-raw-methods.ipynb",
         "dict_head": "03-raw-methods.ipynb",
         "get_available_methods": "03-raw-methods.ipynb",
         "extract_parameter_example": "03-raw-methods.ipynb",
         "construct_parameters": "03-raw-methods.ipynb",
         "construct_path_functions": "03-raw-methods.ipynb",
         "construct_all_functions": "03-raw-methods.ipynb",
         "save_methods": "03-raw-methods.ipynb",
         "test_endpoints": "04-client-prep.ipynb",
         "construct_method_to_params_dict": "04-client-prep.ipynb",
         "construct_request_type_filter": "04-client-prep.ipynb",
         "check_request_type_filter": "04-client-prep.ipynb",
         "determine_request_type_from_fields": "04-client-prep.ipynb",
         "determine_method_request_types": "04-client-prep.ipynb",
         "construct_method_to_params_map": "04-client-prep.ipynb",
         "construct_method_info_dict": "04-client-prep.ipynb",
         "retry_request": "05-orchestrator.ipynb",
         "if_possible_parse_local_datetime": "05-orchestrator.ipynb",
         "SP_and_date_request": "05-orchestrator.ipynb",
         "handle_capping": "05-orchestrator.ipynb",
         "date_range_request": "05-orchestrator.ipynb",
         "year_request": "05-orchestrator.ipynb",
         "construct_year_month_pairs": "05-orchestrator.ipynb",
         "year_and_month_request": "05-orchestrator.ipynb",
         "clean_year_week": "05-orchestrator.ipynb",
         "construct_year_week_pairs": "05-orchestrator.ipynb",
         "year_and_week_request": "05-orchestrator.ipynb",
         "non_temporal_request": "05-orchestrator.ipynb",
         "query_orchestrator": "05-orchestrator.ipynb",
         "generate_streams": "06-client-gen.ipynb",
         "save_api_client": "06-client-gen.ipynb",
         "app": "vis-00-gen.ipynb",
         "rebuild_library": "07-cli-rebuild.ipynb",
         "clean_route_gdf": "09-map-gen.ipynb",
         "load_route_gdf": "09-map-gen.ipynb",
         "construct_df_PN_pivot_dt_rng": "09-map-gen.ipynb",
         "construct_PN_pivot_df": "09-map-gen.ipynb",
         "get_PHYBM_df": "09-map-gen.ipynb",
         "download_latest_PHYBM_data": "09-map-gen.ipynb",
         "get_files": "09-map-gen.ipynb",
         "load_most_recent_PN_data": "09-map-gen.ipynb",
         "construct_osuked_id_mappings": "09-map-gen.ipynb",
         "extract_PN_ts": "09-map-gen.ipynb",
         "construct_map_df": "09-map-gen.ipynb",
         "df_to_gdf": "09-map-gen.ipynb",
         "construct_map_geojson": "09-map-gen.ipynb",
         "save_map_geojson": "09-map-gen.ipynb",
         "get_nearest_dt_idx": "09-map-gen.ipynb",
         "generate_map_js": "09-map-gen.ipynb",
         "generate_map_md": "09-map-gen.ipynb",
         "generate_map": "09-map-gen.ipynb",
         "update_version": "10-nbdev.ipynb",
         "add_init": "10-nbdev.ipynb",
         "export.add_init": "10-nbdev.ipynb",
         "export.update_version": "10-nbdev.ipynb",
         "prepare_nbdev_module": "10-nbdev.ipynb",
         "notebook2script": "10-nbdev.ipynb",
         "add_mod_extra_indices": "10-nbdev.ipynb",
         "add_mod_extra_modules": "10-nbdev.ipynb",
         "add_extra_code_desc_to_mod": "10-nbdev.ipynb",
         "rgb_2_plt_tuple": "vis-01-ei.ipynb",
         "get_vis_func": "vis-00-gen.ipynb",
         "get_vis_md_text": "vis-00-gen.ipynb",
         "get_rerun_vis_bool": "vis-00-gen.ipynb",
         "update_vis_configs": "vis-00-gen.ipynb",
         "update_vis": "vis-00-gen.ipynb",
         "generate_vis": "vis-00-gen.ipynb",
         "get_ei_raw_urls": "vis-01-ei.ipynb",
         "get_ei_df": "vis-01-ei.ipynb",
         "clean_EI_df_for_plot": "vis-01-ei.ipynb",
         "convert_fuel_colour_dict_to_plt_tuple": "vis-01-ei.ipynb",
         "hide_spines": "vis-01-ei.ipynb",
         "stacked_fuel_plot": "vis-01-ei.ipynb",
         "construct_GB_decarb_progess_md_txt": "vis-01-ei.ipynb",
         "generate_GB_decarb_progess": "vis-01-ei.ipynb",
         "get_price_dispatchable_s": "vis-01-ei.ipynb",
         "construct_reg_dates": "vis-01-ei.ipynb",
         "estimate_ts_surface": "vis-01-ei.ipynb",
         "plot_ts_surface_estimate": "vis-01-ei.ipynb",
         "construct_moe_md_txt": "vis-01-ei.ipynb",
         "generate_moe": "vis-01-ei.ipynb",
         "get_wf_ids": "vis-02-curtailment.ipynb",
         "flatten_list": "vis-02-curtailment.ipynb",
         "get_curtailed_wfs_df": "vis-02-curtailment.ipynb",
         "load_curtailed_wfs": "vis-02-curtailment.ipynb",
         "add_next_week_of_data_to_curtailed_wfs": "vis-02-curtailment.ipynb",
         "get_latest_lolpdrm_df": "vis-03-lolp.ipynb",
         "clean_fcst_df": "vis-03-lolp.ipynb",
         "AxTransformer": "vis-03-lolp.ipynb",
         "set_date_ticks": "vis-03-lolp.ipynb",
         "create_fcst_htmp": "vis-03-lolp.ipynb",
         "create_fcst_delta_htmp": "vis-03-lolp.ipynb",
         "save_lolpdrm_imgs": "vis-03-lolp.ipynb",
         "construct_drm_md_txt": "vis-03-lolp.ipynb",
         "generate_lolpdrm_imgs_text": "vis-03-lolp.ipynb",
         "Client": "06-client-gen.ipynb",
         "get_B0610": "03-raw-methods.ipynb",
         "get_B0620": "03-raw-methods.ipynb",
         "get_B0630": "03-raw-methods.ipynb",
         "get_B0640": "03-raw-methods.ipynb",
         "get_B0650": "03-raw-methods.ipynb",
         "get_B0710": "03-raw-methods.ipynb",
         "get_B0720": "03-raw-methods.ipynb",
         "get_B0810": "03-raw-methods.ipynb",
         "get_B0910": "03-raw-methods.ipynb",
         "get_B1010": "03-raw-methods.ipynb",
         "get_B1020": "03-raw-methods.ipynb",
         "get_B1030": "03-raw-methods.ipynb",
         "get_B1320": "03-raw-methods.ipynb",
         "get_B1330": "03-raw-methods.ipynb",
         "get_B1410": "03-raw-methods.ipynb",
         "get_B1420": "03-raw-methods.ipynb",
         "get_B1430": "03-raw-methods.ipynb",
         "get_B1440": "03-raw-methods.ipynb",
         "get_B1510": "03-raw-methods.ipynb",
         "get_B1520": "03-raw-methods.ipynb",
         "get_B1530": "03-raw-methods.ipynb",
         "get_B1540": "03-raw-methods.ipynb",
         "get_B1610": "03-raw-methods.ipynb",
         "get_B1620": "03-raw-methods.ipynb",
         "get_B1630": "03-raw-methods.ipynb",
         "get_B1720": "03-raw-methods.ipynb",
         "get_B1730": "03-raw-methods.ipynb",
         "get_B1740": "03-raw-methods.ipynb",
         "get_B1750": "03-raw-methods.ipynb",
         "get_B1760": "03-raw-methods.ipynb",
         "get_B1770": "03-raw-methods.ipynb",
         "get_B1780": "03-raw-methods.ipynb",
         "get_B1790": "03-raw-methods.ipynb",
         "get_B1810": "03-raw-methods.ipynb",
         "get_B1820": "03-raw-methods.ipynb",
         "get_B1830": "03-raw-methods.ipynb",
         "get_BOD": "03-raw-methods.ipynb",
         "get_CDN": "03-raw-methods.ipynb",
         "get_DERSYSDATA": "03-raw-methods.ipynb",
         "get_DETSYSPRICES": "03-raw-methods.ipynb",
         "get_DEVINDOD": "03-raw-methods.ipynb",
         "get_DISBSAD": "03-raw-methods.ipynb",
         "get_FORDAYDEM": "03-raw-methods.ipynb",
         "get_FREQ": "03-raw-methods.ipynb",
         "get_FUELHH": "03-raw-methods.ipynb",
         "get_LOLPDRM": "03-raw-methods.ipynb",
         "get_MELIMBALNGC": "03-raw-methods.ipynb",
         "get_MID": "03-raw-methods.ipynb",
         "get_MessageDetailRetrieval": "03-raw-methods.ipynb",
         "get_MessageListRetrieval": "03-raw-methods.ipynb",
         "get_NETBSAD": "03-raw-methods.ipynb",
         "get_NONBM": "03-raw-methods.ipynb",
         "get_PHYBMDATA": "03-raw-methods.ipynb",
         "get_ROLSYSDEM": "03-raw-methods.ipynb",
         "get_SYSDEM": "03-raw-methods.ipynb",
         "get_SYSWARN": "03-raw-methods.ipynb",
         "get_TEMP": "03-raw-methods.ipynb",
         "get_WINDFORFUELHH": "03-raw-methods.ipynb"}

modules = ["api.py",
           "dev/clientgen.py",
           "dev/clientprep.py",
           "dev/docs.py",
           "dev/nbdev.py",
           "dev/orchestrator.py",
           "dev/raw.py",
           "dev/rawgen.py",
           "dev/rdf.py",
           "dev/specgen.py",
           "dev/utils.py",
           "rebuild.py",
           "vis/curtailment.py",
           "vis/ei.py",
           "vis/gen.py",
           "vis/lolp.py",
           "vis/map.py"]

doc_url = "https://OSUKED.github.io/ElexonDataPortal/"

git_url = "https://github.com/OSUKED/ElexonDataPortal"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02-spec-gen.ipynb (unless otherwise specified).

__all__ = ['init_spec', 'load_endpoints_df', 'get_endpoint_single_attr', 'init_stream_dict',
           'add_params_to_stream_dict', 'add_streams_to_spec', 'construct_spec', 'save_spec', 'load_API_yaml']

# Cell
import numpy as np
import pandas as pd

import os
import yaml

from jinja2 import Template

# Cell
def init_spec(
    title='BMRS API',
    description='API for the Elexon Balancing Mechanism Reporting Service',
    root_url='https://api.bmreports.com'
):
    API_spec = dict()

    API_spec['title'] = title
    API_spec['description'] = description
    API_spec['root_url'] = root_url

    return API_spec

# Cell
def load_endpoints_df(endpoints_fp: str='data/endpoints.csv'):
    df_endpoints = pd.read_csv(endpoints_fp)

    date_idxs = (df_endpoints['Sample Data'].str.count('/')==2).replace(np.nan, False)
    time_idxs = df_endpoints['Sample Data'].str.contains(':').replace(np.nan, False)

    df_endpoints.loc[date_idxs & ~time_idxs, 'Sample Data'] = pd.to_datetime(df_endpoints.loc[date_idxs & ~time_idxs, 'Sample Data']).dt.strftime('%Y-%m-%d')
    df_endpoints.loc[date_idxs & time_idxs, 'Sample Data'] = pd.to_datetime(df_endpoints.loc[date_idxs & time_idxs, 'Sample Data']).dt.strftime('%Y-%m-%d %H:%M:%S')

    df_endpoints['Sample Data'] = df_endpoints['Sample Data'].fillna('')
    df_endpoints['Field Name'] = df_endpoints['Field Name'].str.replace(' ', '')

    return df_endpoints

# Cell
def get_endpoint_single_attr(df_endpoint, attribute='version'):
    attr_val = df_endpoint[attribute].unique()
    assert len(attr_val)==1, f'Expected only 1 {attribute}, instead found {len(attr_val)}'
    attr_val = attr_val[0]

    return attr_val

# Cell
def init_stream_dict(df_endpoint, endpoint_id):
    version = get_endpoint_single_attr(df_endpoint, 'version')
    name = get_endpoint_single_attr(df_endpoint, 'name')
    tags = get_endpoint_single_attr(df_endpoint, 'tags')

    stream = dict()

    stream['endpoint'] = f'/BMRS/{endpoint_id}/v{version}'
    stream['x-title'] = endpoint_id
    stream['description'] = name
    stream['parameters'] = list()

    if isinstance(tags, str):
        stream['tags'] = tags.split(', ')

    return stream

# Cell
def add_params_to_stream_dict(
    df_endpoint: pd.DataFrame,
    stream: dict,
    field_type_map: dict={
        'String': 'string',
        'Int': 'integer',
        'int': 'integer',
        'Integer': 'integer',
        'Date': 'string'
    }
):
    for _, (param_name, param_type, param_sample) in df_endpoint.query('direction=="request"')[['Field Name', 'Field Type', 'Sample Data']].iterrows():
        parameter = dict()

        parameter['name'] = param_name
        parameter['type'] = field_type_map[param_type]

        if param_type in ['Date']:
            parameter['format'] = 'date'

        if param_name == 'APIKey':
            parameter['format'] = 'password'

        if param_sample == 'csv/xml':
            parameter['examples'] = {f'{param_sub_sample}': {'value': param_sub_sample} for param_sub_sample in param_sample.split('/')}
        else:
            parameter['example'] = param_sample

        stream['parameters'] += [parameter]

    return stream

# Cell
def add_streams_to_spec(API_spec, df_endpoints):
    API_spec['streams'] = list()
    endpoint_ids = sorted(list(df_endpoints['id'].unique()))

    for endpoint_id in endpoint_ids:
        df_endpoint = df_endpoints.query(f'id==@endpoint_id')

        stream = init_stream_dict(df_endpoint, endpoint_id)
        stream = add_params_to_stream_dict(df_endpoint, stream)

        API_spec['streams'] += [stream]

    return API_spec

# Cell
def construct_spec(
    df_endpoints: pd.DataFrame,
    title: str='BMRS API',
    description: str='API for the Elexon Balancing Mechanism Reporting Service',
    root_url: str='https://api.bmreports.com'
):
    API_spec = init_spec()
    API_spec = add_streams_to_spec(API_spec, df_endpoints)

    return API_spec

# Cell
def save_spec(
    API_spec: dict,
    in_fp: str='../templates/open_api_spec.yaml',
    out_fp: str='../data/BMRS_API.yaml'
):
    rendered_schema = Template(open(in_fp).read()).render(API_spec=API_spec)

    with open(out_fp, 'w') as f:
        try:
            f.write(rendered_schema)
        except e as exc:
            raise exc

# Cell
def load_API_yaml(fp='../data/BMRS_API.yaml'):
    with open(fp, 'r') as stream:
        try:
            API_yaml = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            raise exc

    return API_yaml


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/vis-02-curtailment.ipynb (unless otherwise specified).

__all__ = ['get_wf_ids', 'flatten_list', 'get_curtailed_wfs_df', 'load_curtailed_wfs',
           'add_next_week_of_data_to_curtailed_wfs']

# Cell
flatten_list = lambda list_: [item for sublist in list_ for item in sublist]

def get_wf_ids(dictionary_url='https://raw.githubusercontent.com/OSUKED/Power-Station-Dictionary/main/data/output/power_stations.csv'):
    df_dictionary = pd.read_csv(dictionary_url)
    wf_ids = flatten_list(df_dictionary.query('fuel_type=="wind"')['sett_bmu_id'].str.split(', ').to_list())

    return wf_ids

# Cell
def get_curtailed_wfs_df(
    api_key: str=None,
    start_date: str = '2020-01-01',
    end_date: str = '2020-01-01 1:30',
    wf_ids: list=None,
    dictionary_url: str='https://raw.githubusercontent.com/OSUKED/Power-Station-Dictionary/main/data/output/power_stations.csv'
):
    if wf_ids is None:
        wf_ids = get_wf_ids(dictionary_url=dictionary_url)

    if isinstance(start_date, str):
        start_date = pd.to_datetime(start_date).tz_localize('Europe/London')

    if isinstance(end_date, str):
        end_date = pd.to_datetime(end_date).tz_localize('Europe/London')

    client = Client()
    df_detsysprices = client.get_DETSYSPRICES(start_date, end_date)

    df_curtailed_wfs = (df_detsysprices
                        .query('recordType=="BID" & soFlag=="T" & id in @wf_ids')
                        .astype({'bidVolume': float})
                        .groupby(['local_datetime', 'id'])
                        ['bidVolume']
                        .sum()
                        .reset_index()
                        .pivot('local_datetime', 'id', 'bidVolume')
                       )

    df_curtailed_wfs = df_curtailed_wfs.reindex(pd.date_range(min(start_date, df_curtailed_wfs.index.min()),
                                                              max(end_date-pd.Timedelta(minutes=30), df_curtailed_wfs.index.max()),
                                                              tz='Europe/London'))

    df_curtailed_wfs.index.name = 'local_datetime'

    return df_curtailed_wfs

# Cell
def load_curtailed_wfs(
    curtailed_wfs_fp: str='data/curtailed_wfs.csv'
):
    df_curtailed_wfs = pd.read_csv(curtailed_wfs_fp)

    df_curtailed_wfs = df_curtailed_wfs.set_index('local_datetime')
    df_curtailed_wfs.index = pd.to_datetime(df_curtailed_wfs.index, utc=True).tz_convert('Europe/London')

    return df_curtailed_wfs

# Cell
def add_next_week_of_data_to_curtailed_wfs(
    curtailed_wfs_fp: str='data/curtailed_wfs.csv',
    save_data: bool=True,
):
    df_curtailed_wfs = load_curtailed_wfs(curtailed_wfs_fp)
    most_recent_ts = df_curtailed_wfs.index.max()

    start_date = most_recent_ts + pd.Timedelta(minutes=30)
    end_date = start_date + pd.Timedelta(days=7)

    client = Client()
    df_curtailed_wfs_wk = get_curtailed_wfs_df(start_date=start_date, end_date=end_date, wf_ids=wf_ids)

    df_curtailed_wfs = df_curtailed_wfs.append(df_curtailed_wfs_wk)
    df_curtailed_wfs.to_csv(curtailed_wfs_fp)

    return df_curtailed_wfs


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/vis-00-gen.ipynb (unless otherwise specified).

__all__ = ['rgb_2_plt_tuple', 'get_vis_func', 'get_vis_md_text', 'get_rerun_vis_bool', 'update_vis_configs',
           'update_vis', 'generate_vis', 'app']

# Cell
import json
import pandas as pd

import typer
import croniter
import importlib
from tqdm import tqdm

import matplotlib.pyplot as plt

# Cell
def rgb_2_plt_tuple(r, g, b):
    """converts a standard rgb set from a 0-255 range to 0-1"""

    plt_tuple = tuple([x/255 for x in (r, g, b)])

    return plt_tuple

# Cell
def get_vis_func(func_path):
    *lib_path, func_name = func_path.split('.')
    lib_obj = importlib.import_module('.'.join(lib_path))

    func = getattr(lib_obj, func_name)

    return func

def get_vis_md_text(vis_config, docs_dir=None, update_time=None):
    func_path = vis_config['function']
    kwargs = vis_config['kwargs']

    if (docs_dir is not None) and ('docs_dir' in kwargs.keys()):
        kwargs['docs_dir'] = docs_dir

    if (update_time is not None) and ('update_time' in kwargs.keys()):
        kwargs['update_time'] = update_time

    vis_func = get_vis_func(func_path)
    vis_md_text = vis_func(**kwargs)
    plt.close()

    return vis_md_text

# Cell
def get_rerun_vis_bool(vis_config):
    if 'last_update_time' not in vis_config.keys():
        return True
    else:
        last_update_time = pd.to_datetime(vis_config['last_update_time']).tz_localize('Europe/London')

    cron = croniter.croniter(vis_config['cron'], pd.Timestamp.now()-pd.Timedelta(weeks=1))
    cron_dts = pd.to_datetime([cron.get_next() for i in range(10*48*7)], unit='s').tz_localize('UTC').tz_convert('Europe/London')

    s_cron_dts_time_delta_to_now = pd.Series((cron_dts - pd.Timestamp.now(tz='Europe/London')).total_seconds())
    assert (s_cron_dts_time_delta_to_now<0).sum()>0 and (s_cron_dts_time_delta_to_now>0).sum()>0, 'The cron dts being assessed do not cross the current time'

    s_cron_dts_time_delta_to_last_update_time = pd.Series((cron_dts - last_update_time).total_seconds())
    if s_cron_dts_time_delta_to_now.abs().idxmin() == s_cron_dts_time_delta_to_last_update_time.abs().idxmin():
        return False

    avg_adj_dt_time_delta_s = pd.Series(cron_dts).diff(1).dropna().dt.total_seconds().mean()
    min_time_delta_s = s_cron_dts_time_delta_to_now.abs().min()

    rerun_vis = avg_adj_dt_time_delta_s >= min_time_delta_s

    return rerun_vis

# Cell
def update_vis_configs(
    vis_configs,
    docs_dir: str='docs',
    override_rerun_vis_bool: bool=False
):
    for i, vis_config in enumerate(vis_configs):
        update_time = pd.Timestamp.now().round('5min').strftime('%Y-%m-%d %H:%M')
        rerun_vis = get_rerun_vis_bool(vis_config)

        if override_rerun_vis_bool == True:
            rerun_vis = True

        if rerun_vis == True:
            vis_md_text = get_vis_md_text(vis_config, docs_dir=docs_dir, update_time=update_time)
            vis_configs[i]['md_text'] = vis_md_text
            vis_configs[i]['last_update_time'] = update_time

    return vis_configs

# Cell
app = typer.Typer()

@app.command()
def update_vis(
    docs_dir: str='docs',
    data_dir: str='data',
    override_rerun_vis_bool: bool=False
):
    with open(f'{data_dir}/vis_configs.json', 'r') as f:
        vis_configs = json.load(f)

    vis_configs = update_vis_configs(vis_configs, docs_dir=docs_dir, override_rerun_vis_bool=override_rerun_vis_bool)

    with open(f'{data_dir}/vis_configs.json', 'w') as f:
        json.dump(vis_configs, f)

    prefix_text = """# Visualisations

On this page you can view visualisations of key phenomena in the GB power sector, ranging from long-term trends in the generation-mix and market prices to information on excess capacity in the grid. All data used in these visualisations was either sourced directly from BMRS using the `ElexonDataPortal` client, or has been derived from BMRS data streams. As with the other components of the `ElexonDataPortal` the code to generate these visualisations is open-source and users are welcome to contribute their own visualisations, for more detail on how to do this please refer to the [user contribution guide](#contributor-guide)
    """

    suffix_text = """### Contributor Guide

We encourage users to contribute their own visualisations which the `ElexonDataPortal` will then update automatically. To this end the library adopts a standardised format for generating visualisations, the core component of which is the `data/vis_configs.json` file to which you will have to add detail on your visualisation function:

```javascript
[
    ...
    {
        "cron": "0 * * * *", # the update schedule, in this instance to run at midnight every sunday
        "function": "path_to_function", # e.g. ElexonDataPortal.vis.generate_vis
        "kwargs": {
            'api_key': null,  # if no api_key is passed then the client will try and look for the `BMRS_API_KEY` environment variable
            'update_time': null, # if no update_time is passed you should generate it yourself, e.g. with `pd.Timestamp.now().round('5min').strftime('%Y-%m-%d %H:%M')`
            'docs_dir': 'docs', # in almost all circumstances this should just be `docs`
            "optional_kwarg": "optional_value" # you can specify any additional keyword arguments that your function requires
        }
    },
    ...
]
```

<br>

The other core component is writing the function that generates the visualisation. This function should require parameters for the `docs_dir`, `api_key`, and `update_time` but can include optional parameters that you wish to specify, it should then return markdown text which will be used to populate the *Visualisations* page. These functions will normally contain three steps: data retrieval, generating the visualisation, and generating the accompanying text - an example can be seen below.

```python
import pandas as pd
import matplotlib.pyplot as plt
from ..api import Client

def generate_vis(
    docs_dir: str='docs',
    api_key: str=None,
    update_time: str=pd.Timestamp.now().round('5min').strftime('%Y-%m-%d %H:%M'),
) -> str:

    # Data Retrieval
    client = Client(api_key=api_key)
    df = client.get_data_stream(param1, param2)

    # Generating the Visualisation
    fig, ax = plt.subplots(dpi=150)
    df.plot(ax=ax)
    fig.savefig(f'{docs_dir}/img/vis/great_vis_name.png')

    # Generating the Text
    md_text = f\"\"\"### Title

Explanation of what your visualisation shows

![](img/vis/great_vis_name.png)
\"\"\"

    return md_text
```

N.b. the path to the image should be relative to the `docs` directory.

If you require any assistance in this process please start a discussion [here](https://github.com/OSUKED/ElexonDataPortal/discussions) and we'll endeavour to help as best we can.
"""

    all_vis_md_texts = [prefix_text] + [vis_config['md_text'] for vis_config in vis_configs] + [suffix_text]
    combined_md_text = '\n\n<br>\n\n'.join(all_vis_md_texts)

    with open(f'{docs_dir}/visualisations.md', 'w', encoding='utf-8') as f:
        f.write(combined_md_text)

    return

if __name__ == '__main__' and '__file__' in globals():
    app()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/vis-03-lolp.ipynb (unless otherwise specified).

__all__ = ['get_latest_lolpdrm_df', 'clean_fcst_df', 'AxTransformer', 'set_date_ticks', 'create_fcst_htmp',
           'create_fcst_delta_htmp', 'save_lolpdrm_imgs', 'construct_drm_md_txt', 'generate_lolpdrm_imgs_text']

# Cell
import numpy as np
import pandas as pd

import seaborn as sns
import matplotlib.pyplot as plt

from collections.abc import Iterable
from sklearn import linear_model

from ..api import Client

# Cell
def get_latest_lolpdrm_df(api_key: str=None):
    start_date = pd.Timestamp.now(tz='Europe/London').strftime('%Y-%m-%d')
    end_date = pd.Timestamp.now(tz='Europe/London') + pd.Timedelta(days=3)

    client = Client(api_key=api_key)
    df_lolpdrm = client.get_LOLPDRM(start_date, end_date)

    return df_lolpdrm

# Cell
def clean_fcst_df(
    df_lolpdrm: pd.DataFrame,
    fcst_type: str='lolp',
    fcst_horizons: list=[8, 4, 2, 1]
):
    cols_renaming_map = {f'{fcst_type}12Forecast': '12:00'}

    cols_renaming_map.update({
        f'{fcst_type}{fcst_horizon}HourForecast': f"{fcst_horizon} Hour{'s' if fcst_horizon>1 else ''} Ahead"
        for fcst_horizon
        in fcst_horizons
    })

    df_clean_fcst = (df_lolpdrm
                     .set_index('local_datetime')
                     [cols_renaming_map.keys()]
                     .rename(columns=cols_renaming_map)
                     .astype(float)
                    )

    return df_clean_fcst

# Cell
class AxTransformer:
    def __init__(self, datetime_vals=False):
        self.datetime_vals = datetime_vals
        self.lr = linear_model.LinearRegression()

        return

    def process_tick_vals(self, tick_vals):
        if not isinstance(tick_vals, Iterable) or isinstance(tick_vals, str):
            tick_vals = [tick_vals]

        if self.datetime_vals == True:
            tick_vals = pd.to_datetime(tick_vals).astype(int).values

        tick_vals = np.array(tick_vals)

        return tick_vals

    def fit(self, ax, axis='x'):
        axis = getattr(ax, f'get_{axis}axis')()

        tick_locs = axis.get_ticklocs()
        tick_vals = self.process_tick_vals([label._text for label in axis.get_ticklabels()])

        self.lr.fit(tick_vals.reshape(-1, 1), tick_locs)

        return

    def transform(self, tick_vals):
        tick_vals = self.process_tick_vals(tick_vals)
        tick_locs = self.lr.predict(np.array(tick_vals).reshape(-1, 1))

        return tick_locs

def set_date_ticks(ax, start_date, end_date, axis='y', date_format='%Y-%m-%d', **date_range_kwargs):
    dt_rng = pd.date_range(start_date, end_date, **date_range_kwargs)

    ax_transformer = AxTransformer(datetime_vals=True)
    ax_transformer.fit(ax, axis=axis)

    getattr(ax, f'set_{axis}ticks')(ax_transformer.transform(dt_rng))
    getattr(ax, f'set_{axis}ticklabels')(dt_rng.strftime(date_format))

    ax.tick_params(axis=axis, which='both', bottom=True, top=False, labelbottom=True)

    return ax

# Cell
def create_fcst_htmp(
    df_clean: pd.DataFrame,
    img_fp: str='docs/img/vis/drm_fcst_htmp.png'
):
    # Plotting
    fig, ax = plt.subplots(dpi=250)

    htmp = sns.heatmap(df_clean, vmin=0, ax=ax)

    set_date_ticks(ax, df_clean.index.min(), df_clean.index.max(), freq='6H', date_format='%Y-%m-%d %H:%M')
    ax.set_xticklabels(ax.get_xticklabels(), rotation=25)
    ax.xaxis.tick_top()
    ax.set_ylabel('')
    cbar = htmp.collections[0].colorbar
    cbar.set_label('De-Rated Margin (MW)', labelpad=20, rotation=270)

    # Saving
    fig.tight_layout()
    fig.savefig(img_fp)

# Cell
def create_fcst_delta_htmp(
    df_clean_drm: pd.DataFrame,
    img_fp: str='docs/img/vis/drm_fcst_delta_htmp.png'
):
    # Preparing data
    df_drm_delta = df_clean_drm.iloc[:, 2:].apply(lambda col: col-df_clean_drm.iloc[:, 1]).dropna(how='all', axis=1).dropna(how='all')

    # Plotting
    fig, ax = plt.subplots(dpi=150)

    htmp = sns.heatmap(df_drm_delta, center=0, cmap='bwr_r', ax=ax)

    set_date_ticks(ax, df_drm_delta.index.min(), df_drm_delta.index.max(), freq='3H', date_format='%Y-%m-%d %H:%M')
    ax.set_xticklabels(ax.get_xticklabels(), rotation=25)
    ax.xaxis.tick_top()
    ax.set_ylabel('')
    cbar = htmp.collections[0].colorbar
    cbar.set_label('Delta to the 8 Hour-\nAhead Forecast Horizon', labelpad=25, rotation=270)

    # Saving
    fig.tight_layout()
    fig.savefig(img_fp)

# Cell
def save_lolpdrm_imgs(
    docs_dir: str='docs',
    api_key: str=None,
    fcst_type: str='lolp',
    fcst_horizons: list=[8, 4, 2, 1]
):
    df_lolpdrm = get_latest_lolpdrm_df(api_key=api_key)
    df_clean = clean_fcst_df(df_lolpdrm, fcst_type=fcst_type, fcst_horizons=fcst_horizons)

    create_fcst_htmp(df_clean, img_fp=f'{docs_dir}/img/vis/{fcst_type}_fcst_htmp.png')
    plt.close()

    create_fcst_delta_htmp(df_clean, img_fp=f'{docs_dir}/img/vis/{fcst_type}_fcst_delta_htmp.png')
    plt.close()

# Cell
def construct_drm_md_txt(
    update_time: str=None
):
    if update_time is None:
        update_time = pd.Timestamp.now().round('5min').strftime('%Y-%m-%d %H:%M')

    md_txt = f"""### De-Rated Margin

In each settlement period the system operator publishes the de-rated margin forecast calculated in accordance with the [Loss of Load Probability Calculation Statement](https://www.elexon.co.uk/documents/bsc-codes/lolp/loss-of-load-probability-calculation-statement/) at the following times:

* At 1200 hours on each calendar day for all Settlement Periods for which Gate Closure has not yet passed and which occur within the current Operational Day or the following Operational Day; and
* At eight, four, two and one hour(s) prior to the beginning of the Settlement Period to which the De-Rated Margin Forecast relates.

These figures will be updated on an hourly basis, the last update was at: {update_time}

<br>

#### Forecasts

The following heatmap shows the evolving de-rated margin forecast across the different forecast horizons.

![](img/vis/drm_fcst_htmp.png)

<br>

#### Forecast Deltas

The following heatmap shows how the more recent de-rated margin forecasts deviate from the 8 hours ahead forecast.

![](img/vis/drm_fcst_delta_htmp.png)
"""

    return md_txt

# Cell
def generate_lolpdrm_imgs_text(
    docs_dir: str='docs',
    api_key: str=None,
    fcst_horizons: list=[8, 4, 2, 1],
    update_time: str=None
):
    for fcst_type in ['drm', 'lolp']:
        save_lolpdrm_imgs(docs_dir=docs_dir, api_key=api_key, fcst_type=fcst_type, fcst_horizons=fcst_horizons)

    md_txt = construct_drm_md_txt(update_time=update_time)

    return md_txt


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/09-map-gen.ipynb (unless otherwise specified).

__all__ = ['clean_route_gdf', 'load_route_gdf', 'construct_df_PN_pivot_dt_rng', 'construct_PN_pivot_df', 'get_PHYBM_df',
           'download_latest_PHYBM_data', 'get_files', 'load_most_recent_PN_data', 'construct_osuked_id_mappings',
           'extract_PN_ts', 'construct_map_df', 'df_to_gdf', 'construct_map_geojson', 'save_map_geojson',
           'get_nearest_dt_idx', 'generate_map_js', 'generate_map_md', 'generate_map']

# Cell
import json
import numpy as np
import pandas as pd
import geopandas as gpd

import os
from tqdm import tqdm
from jinja2 import Template
from warnings import warn

from ..dev import utils, raw

# Cell
def clean_route_gdf(gdf):
    s_LV_routes = gdf['OPERATING_'].astype(float).fillna(0)<=132
    gdf.loc[s_LV_routes, 'OPERATING_'] = '<=132'
    gdf.loc[~s_LV_routes, 'OPERATING_'] = gdf.loc[~s_LV_routes, 'OPERATING_'].astype(int).astype(str)

    return gdf

def load_route_gdf(data_dir='data'):
    gdf_OHL = gpd.read_file(f'{data_dir}/transmission-system/OHL.shp').to_crs('EPSG:4326')
    gdf_cables = gpd.read_file(f'{data_dir}/transmission-system/Cable.shp').to_crs('EPSG:4326')

    gdf_OHL = clean_route_gdf(gdf_OHL)
    gdf_cables = clean_route_gdf(gdf_cables)

    gdf_route = gdf_OHL.append(gdf_cables)[['OPERATING_', 'geometry']].rename(columns={'OPERATING_': 'kV'})
    gdf_route.to_file(f'{data_dir}/network_routes.json', driver='GeoJSON')

    return gdf_route

# Cell
def construct_df_PN_pivot_dt_rng(df_PN):
    no_seconds = (((df_PN['timeFrom'].str.split(':').str[-1]=='00').mean()==1) &
                  ((df_PN['timeTo'].str.split(':').str[-1]=='00').mean()==1))

    if no_seconds == True:
        dt_rng = pd.date_range(df_PN['timeFrom'].min(), df_PN['timeTo'].max(), freq='min', tz='Europe/London')
    else:
        dt_rng = pd.date_range(df_PN['timeFrom'].min(), df_PN['timeTo'].max(), freq='s', tz='Europe/London')

    return dt_rng

def construct_PN_pivot_df(df_PN, resample=None):
    bmu_ids = sorted(list(df_PN['bmUnitID'].unique()))
    df_PN_pivot = pd.DataFrame(index=construct_df_PN_pivot_dt_rng(df_PN), columns=bmu_ids, dtype=float)

    for bmu_id in tqdm(bmu_ids):
        for idx, row in df_PN.query('bmUnitID==@bmu_id').iterrows():
            df_PN_pivot.loc[pd.to_datetime(row['timeFrom']).tz_localize('Europe/London'), bmu_id] = float(row['pnLevelFrom'])
            df_PN_pivot.loc[pd.to_datetime(row['timeTo']).tz_localize('Europe/London'), bmu_id] = float(row['pnLevelTo'])

        df_PN_pivot[bmu_id] = df_PN_pivot[bmu_id].interpolate()

    if resample is not None:
        df_PN_pivot = df_PN_pivot.resample(resample).mean()

    return df_PN_pivot

def get_PHYBM_df(api_key, start_date=None, end_date=None, record_type='PN', resample='30T'):
    if start_date is None and end_date is None:
        start_date = pd.Timestamp.now().round('30min') - pd.Timedelta(minutes=60*1)
        end_date = pd.Timestamp.now().round('30min') + pd.Timedelta(minutes=180)
    elif start_date is not None and end_date is not None:
        pass
    else:
        raise ValueError('Only one of `start_date` and `end_date` was specified')

    df = pd.DataFrame()
    df_local_datetime_to_date_SP = utils.dt_rng_to_SPs(start_date, end_date)

    for idx, (date, SP) in tqdm(df_local_datetime_to_date_SP.iterrows(), total=df_local_datetime_to_date_SP.shape[0]):
        df_SP = utils.parse_xml_response(raw.get_PHYBMDATA(api_key, SettlementDate=date, SettlementPeriod=SP, ServiceType='xml'))
        df = df.append(df_SP)

    df = (df
          .query(f'recordType=="{record_type}"')
          .dropna(how='all', axis=1)
         )

    if resample is not None:
        df = df.pipe(construct_PN_pivot_df, resample=resample)

    df.index.name = 'local_datetime'

    return df

# Cell
get_files = lambda data_dir: [f for f in os.listdir(data_dir) if '.csv' in f]

def download_latest_PHYBM_data(api_key=None, data_dir='data/PN', record_type='PN'):
    if api_key is None:
        if 'BMRS_API_KEY' in os.environ.keys():
            api_key = os.environ['BMRS_API_KEY']
        else:
            raise ValueError('`api_key` must be passed or set as the environment variable `BMRS_API_KEY`')

    files = get_files(data_dir)
    years_months_downloaded = [f.split('.')[0] for f in files]

    current_ts = pd.Timestamp.now(tz='Europe/London')
    current_year_month = current_ts.strftime('%Y-%m')

    if current_year_month not in years_months_downloaded:
        start_date, end_date = f'{current_year_month}-01 00:00', current_ts.strftime('%Y-%m-%d %H:%M')
        df = get_PHYBM_df(api_key, start_date, end_date, record_type=record_type)
        df.to_csv(f'{data_dir}/{current_year_month}.csv')

    else:
        df = pd.read_csv(f'{data_dir}/{current_year_month}.csv')

        df = df.set_index('local_datetime')
        df.index = pd.to_datetime(df.index, utc=True).tz_convert('Europe/London')
        dt_rng = pd.date_range(df.index.max(), current_ts, freq='30T', tz='Europe/London')

        if dt_rng.size > 1:
            start_date = dt_rng[0] - pd.Timedelta(minutes=30)
            end_date = dt_rng[-1] + pd.Timedelta(minutes=60)

            try:
                df_latest = get_PHYBM_df(api_key, start_date, end_date, record_type=record_type)
                df_trimmed = df.drop(list(set(df_latest.index) - (set(df_latest.index) - set(df.index))))
                df_combined = df_trimmed.append(df_latest).sort_index()
                df_combined.to_csv(f'{data_dir}/{current_year_month}.csv')
            except:
                warn(f'Could not retrieve any new data between {start_date} and {end_date}')

# Cell
def load_most_recent_PN_data(data_dir='data/PN', latest_year_month_file=None, df_PN_old=None):
    if latest_year_month_file is None:
        PN_files = sorted(get_files(data_dir))
        latest_year_month_file = PN_files[-1]

    latest_fp = f'{data_dir}/{latest_year_month_file}'

    df_PN = pd.read_csv(latest_fp)

    df_PN['local_datetime'] = pd.to_datetime(df_PN['local_datetime'], utc=True)
    df_PN = df_PN.set_index('local_datetime')
    df_PN.index = df_PN.index.tz_convert('Europe/London')

    if df_PN_old is not None:
        assert latest_year_month_file is not None, 'Should not be appending to the main dataframe if `latest_year_month_file` was not specified'
        df_PN = df_PN_old.append(df_PN).pipe(lambda df: df[~df.index.duplicated(keep='first')]).sort_index()

    if df_PN.shape[0] < (48*7): # want to have at least a week's worth of data
        assert 'PN_files' in locals(), 'The two most recent files combined have less than one week\'s worth of data'
        df_PN = load_most_recent_PN_data(data_dir, latest_year_month_file=PN_files[-2], df_PN_old=df_PN)

    return df_PN

# Cell
def construct_osuked_id_mappings(df_powerdict):
    osuked_id_mappings = dict()

    osuked_id_mappings['bmu_ids'] = (df_powerdict
                                     .set_index('osuked_id')
                                     ['sett_bmu_id']
                                     .str.split(', ')
                                     .dropna()
                                     .to_dict()
                                    )

    osuked_id_mappings['capacity_mw'] = (df_powerdict
                                         .set_index('osuked_id')
                                         ['capacity_mw']
                                         .astype(str)
                                         .str.replace('.0', '', regex=False)
                                         .to_dict()
                                        )

    osuked_id_mappings['fuel_type'] = (df_powerdict
                                       .set_index('osuked_id')
                                       ['fuel_type']
                                       .dropna()
                                       .to_dict()
                                      )

    osuked_id_mappings['name'] = (df_powerdict
                                  .set_index('osuked_id')
                                  ['name']
                                  .dropna()
                                  .to_dict()
                                 )

    osuked_id_mappings['lat_lon'] = (df_powerdict
                                     .set_index('osuked_id')
                                     [['latitude', 'longitude']]
                                     .dropna()
                                     .apply(dict, axis=1)
                                     .to_dict()
                                    )

    return osuked_id_mappings

# Cell
def extract_PN_ts(df_PN, bmu_ids, n_SPs=48*7):
    matching_output_bmu_ids = df_PN.columns.intersection(bmu_ids)
    output_match = matching_output_bmu_ids.size > 0

    if output_match == False:
        return None

    s_PN = df_PN[matching_output_bmu_ids].sum(axis=1)
    s_PN.index = (s_PN.index.tz_convert('UTC') - pd.to_datetime(0, unit='s').tz_localize('UTC')).total_seconds().astype(int) * 1000

    s_PN = s_PN.fillna(0)
    s_PN[s_PN<0] = 0
    PN_ts = s_PN.tail(n_SPs).to_dict()

    return PN_ts

def construct_map_df(
    df_PN,
    osuked_id_to_bmu_ids,
    osuked_id_to_capacity_mw,
    osuked_id_to_lat_lon,
    osuked_id_to_fuel_type,
    osuked_id_to_name,
    n_SPs=48*7
):
    sites_data = list()

    for osuked_id, bmu_ids in osuked_id_to_bmu_ids.items():
        lat_lon_match = osuked_id in osuked_id_to_lat_lon.keys()

        PN_ts = extract_PN_ts(df_PN, bmu_ids, n_SPs=n_SPs)

        if lat_lon_match and PN_ts is not None:
            if sum(PN_ts.values()) > 0:
                site_data = osuked_id_to_lat_lon[osuked_id]
                site_data.update({'id': osuked_id})
                site_data.update({'name': osuked_id_to_name[osuked_id]})
                site_data.update({'capacity': osuked_id_to_capacity_mw[osuked_id]})
                site_data.update({'fuel_type': osuked_id_to_fuel_type[osuked_id]})
                site_data.update({'output': PN_ts})

                sites_data += [site_data]

    df_map = pd.DataFrame(sites_data).set_index('id')

    return df_map

# Cell
def df_to_gdf(df, lat_col='latitude', lon_col='longitude'):
    geometry = gpd.points_from_xy(df[lon_col], df[lat_col])
    gdf = gpd.GeoDataFrame(df, geometry=geometry)

    return gdf

# Cell
def construct_map_geojson(
    df_PN,
    df_powerdict,
    n_SPs=48*7
):
    osuked_id_mappings = construct_osuked_id_mappings(df_powerdict)
    osuked_id_to_bmu_ids, osuked_id_to_capacity_mw, osuked_id_to_fuel_type, osuked_id_to_name, osuked_id_to_lat_lon = osuked_id_mappings.values()

    df_map = construct_map_df(df_PN, osuked_id_to_bmu_ids, osuked_id_to_capacity_mw, osuked_id_to_lat_lon, osuked_id_to_fuel_type, osuked_id_to_name, n_SPs=n_SPs)
    gdf_map = df_to_gdf(df_map)

    geojson = json.loads(gdf_map.to_json().replace('"nan"', 'null'))
    geojson['timeseries'] = [int(unix_datetime) for unix_datetime in list(geojson['features'][0]['properties']['output'].keys())]

    return geojson

# Cell
def save_map_geojson(geojson, fp='data/power_plants.json'):
    with open(fp, 'w') as f:
        json.dump(geojson, f)

# Cell
def get_nearest_dt_idx(geojson, nearest_half_hour):
    ts = pd.to_datetime([x*1000000 for x in geojson['timeseries']]).tz_localize('UTC').tz_convert('Europe/London')
    nearest_hh_match = [i for i, dt in enumerate(ts) if dt==nearest_half_hour]

    if len(nearest_hh_match) == 1:
        return nearest_hh_match[0]
    else:
        return len(ts)-1

def generate_map_js(
    df_PN: pd.DataFrame,
    df_powerdict: pd.DataFrame,
    zoom: int=5,
    center: list=[54.8, -4.61],
    js_template_fp: str='templates/map.js',
    js_docs_fp: str='docs/js/map.js',
    plants_geojson_fp: str='data/power_plants.json',
    plants_geojson_url: str='https://raw.githubusercontent.com/OSUKED/ElexonDataPortal/master/data/power_plants.json',
    routes_geojson_url: str='https://raw.githubusercontent.com/OSUKED/ElexonDataPortal/master/data/network_routes.json'
):
    geojson = construct_map_geojson(df_PN, df_powerdict)
    save_map_geojson(geojson, fp=plants_geojson_fp)

    nearest_half_hour = (pd.Timestamp.now().tz_localize('Europe/London')+pd.Timedelta(minutes=15)).round('30min')
    nearest_dt_idx = get_nearest_dt_idx(geojson, nearest_half_hour)

    rendered_map_js = Template(open(js_template_fp).read()).render(
        zoom=zoom,
        start_idx=nearest_dt_idx,
        center=center,
        plants_geojson_url=plants_geojson_url,
        routes_geojson_url=routes_geojson_url,
    )

    with open(js_docs_fp, 'w', encoding='utf8') as fp:
        fp.write(rendered_map_js)

    return

# Cell
def generate_map_md(md_template_fp='templates/map.md', update_time=None):
    if update_time is None:
        update_time = pd.Timestamp.now().round('5min').tz_localize('Europe/London').strftime('%Y-%m-%d %H:%M')

    rendered_map_md = Template(open(md_template_fp).read()).render({'update_time': update_time})

    return rendered_map_md

# Cell
def generate_map(
    data_dir: str='data/PN',
    api_key: str=None,
    update_time: str=None,
    powerdict_url: str='https://raw.githubusercontent.com/OSUKED/Power-Station-Dictionary/main/data/output/power_stations.csv',
    js_template_fp: str='templates/map.js',
    js_docs_fp: str='docs/js/map.js',
    md_template_fp: str='templates/map.md',
    plants_geojson_fp: str='data/power_plants.json',
    plants_geojson_url: str='https://raw.githubusercontent.com/OSUKED/ElexonDataPortal/master/data/power_plants.json',
    routes_geojson_url: str='https://raw.githubusercontent.com/OSUKED/ElexonDataPortal/master/data/network_routes.json'
):
    if api_key is None:
        assert 'BMRS_API_KEY' in os.environ.keys(), 'If the `api_key` is not specified during client initialisation then it must be set to as the environment variable `BMRS_API_KEY`'
        api_key = os.environ['BMRS_API_KEY']

    download_latest_PHYBM_data(api_key, data_dir)
    df_PN = load_most_recent_PN_data(data_dir)

    df_powerdict = pd.read_csv(powerdict_url)

    generate_map_js(df_PN, df_powerdict, js_template_fp=js_template_fp, js_docs_fp=js_docs_fp, plants_geojson_fp=plants_geojson_fp, plants_geojson_url=plants_geojson_url, routes_geojson_url=routes_geojson_url)
    rendered_map_md = generate_map_md(md_template_fp=md_template_fp, update_time=update_time)

    return rendered_map_md



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/11-rdf.ipynb (unless otherwise specified).

__all__ = []

# Cell
import yaml
import requests
import matplotlib.pyplot as plt

from datacatalogtordf import Catalog
from oastodcat import OASDataService

import networkx as nx
from rdflib import Graph
from rdflib.extras.external_graph_libs import rdflib_to_networkx_multidigraph


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01-utils.ipynb (unless otherwise specified).

__all__ = ['RequestError', 'check_status', 'check_capping', 'expand_cols', 'parse_xml_response', 'dt_rng_to_SPs',
           'parse_local_datetime']

# Cell
import numpy as np
import pandas as pd

import re
import os
import xmltodict
from collections import OrderedDict
from warnings import warn

# Cell
class RequestError(Exception):
    def __init__(self, http_code, error_type, description):
        self.message = f'{http_code} - {error_type}\n{description}'

    def __str__(self):
        return self.message

# Cell
def check_status(r):
    r_metadata = xmltodict.parse(r.text)['response']['responseMetadata']

    if r_metadata['httpCode'] == '204':
        warn(f'Data request was succesful but no content was returned')
        return pd.DataFrame()

    elif r_metadata['httpCode'] != '200':
        raise RequestError(r_metadata['httpCode'], r_metadata['errorType'], r_metadata['description'])

    return None

def check_capping(r):
    r_metadata = xmltodict.parse(r.text)['response']['responseMetadata']

    if 'cappingApplied' in r_metadata.keys():
        if r_metadata['cappingApplied'] == 'Yes':
            capping_applied = True
        else:
            capping_applied = False
    else:
        capping_applied = 'Could not be determined'

    return capping_applied

# Cell
def expand_cols(df, cols_2_expand=[]):
    if df.size == 0:
        return df

    for col in cols_2_expand:
        new_df_cols = df[col].apply(pd.Series)

        df[new_df_cols.columns] = new_df_cols
        df = df.drop(columns=col)

    s_cols_2_expand = df.iloc[0].apply(type).isin([OrderedDict, dict, list, tuple])

    if s_cols_2_expand.sum() > 0:
        cols_2_expand = s_cols_2_expand[s_cols_2_expand].index
        df = expand_cols(df, cols_2_expand)

    return df

def parse_xml_response(r):
    r_dict = xmltodict.parse(r.text)

    status_check_response = check_status(r)
    if status_check_response is not None:
        return status_check_response

    capping_applied = check_capping(r)

    data_content = r_dict['response']['responseBody']['responseList']['item']

    if isinstance(data_content, list):
        df = expand_cols(pd.DataFrame(data_content))
    elif isinstance(data_content, OrderedDict) or isinstance(data_content, dict):
        df = pd.DataFrame(pd.Series(data_content)).T
    else:
        raise ValueError('The returned `data_content` must be one of: `list`, `dict`, or `OrderedDict`')

    return df

# Cell
def dt_rng_to_SPs(
    start_date: pd.Timestamp,
    end_date: pd.Timestamp,
    freq: str='30T',
    tz: str='UTC'
):
    extended_start_date = pd.to_datetime(start_date) - pd.Timedelta(days=1)
    extended_end_date = pd.to_datetime(end_date) + pd.Timedelta(days=1)
    
    dt_rng = pd.date_range(extended_start_date, extended_end_date, freq=freq, tz=tz)

    SPs = list((2*(dt_rng.hour + dt_rng.minute/60) + 1).astype(int))
    dt_strs = list(dt_rng.strftime('%Y-%m-%d'))

    df_dates_SPs = pd.DataFrame({'date':dt_strs, 'SP':SPs}, index=dt_rng).astype(str)

    # Accounting for clock changes
    clock_change_dt_idxs_dir = pd.Series(dt_rng).apply(lambda dt: dt.utcoffset().total_seconds()).diff().replace(0, np.nan).dropna()

    for dt_idx, dir_ in clock_change_dt_idxs_dir.items():
        dt = dt_rng[dt_idx].date()
        SPs = (1 + 2*(dt_rng[dt_rng.date==dt] - pd.to_datetime(dt).tz_localize('Europe/London')).total_seconds()/(60*60)).astype(int)

        df_dates_SPs.loc[df_dates_SPs.index.date==dt, 'SP'] = SPs
        
    tz_label = tz.lower().replace('/', '_')
    df_dates_SPs.index.name = f'datetime_{tz_label}'
    df_dates_SPs['SP'] = df_dates_SPs['SP'].astype(int)
    
    df_dates_SPs = df_dates_SPs[start_date:end_date]

    return df_dates_SPs

def parse_local_datetime(
    df: pd.DataFrame,
    dt_col: str='settlementDate',
    SP_col: str='settlementPeriod',
    freq: str='30T',
    tz: str='Europe/London'
) -> pd.DataFrame:
    # preparing start/end dates
    start_date = pd.to_datetime(df[dt_col].min()) - pd.Timedelta(days=2)
    end_date = pd.to_datetime(df[dt_col].max()) + pd.Timedelta(days=2)

    # mapping from date and SP to local datetime
    df_dates_SPs = dt_rng_to_SPs(start_date, end_date, freq=freq, tz=tz)
    date_SP_to_ts = {(v[0], str(v[1])): k for k, v in df_dates_SPs.apply(tuple, axis=1).to_dict().items()}

    df['local_datetime'] = df[[dt_col, SP_col]].apply(tuple, axis=1).map(date_SP_to_ts)

    # reordering the `local_datetime` column to be first
    cols = list(df.columns)
    cols.remove('local_datetime')
    df = df[['local_datetime'] + cols]

    return df
